/**
  Vertex buffer object demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2008-02-14 (Public Domain)
*/
#include <GL/glew.h> /*<- for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <GL/glut.h>
#include <iostream>
#include <vector>
#include <math.h>
#include "osl/vec4.h" /*<- handy little vector class */
/* Convert a struct with .x, .y, .z fields into three arguments */
#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z

/*********************** App Code **********************/

/* Class representing an OpenGL vertex */
class myVertex {
public:
	vec3 loc; /* 3D location of vertex */
	vec3 color; /* color of vertex */
	myVertex(vec3 l,vec3 n) :loc(l), color(n) {}
};

/* Class representing an OpenGL vertex buffer object */
class myMesh {
public:
	GLuint vb; /* OpenGL "handle" to a vertex buffer object */
	int len; /* Number of elements in our vertex buffer object */
	
	/* Set up an on-graphics-card buffer pointing to this vertex data */
	myMesh(const myVertex *v,int n);
	
	/* Draw the vertices of this mesh */
	void draw(GLenum mode=GL_TRIANGLES);
	
	/* Destroy the graphics card's copy of our data */
	~myMesh();
};

/* Return the distance, in bytes, between these two pointers */
int ptrdiff(const void *a,const void *b) {
	return ((const char *)b)-((const char *)a);
}

/* Set up an on-graphics-card buffer pointing to this vertex data */
myMesh::myMesh(const myVertex *v,int n)
	:vb(0), len(n)
{
	glGenBuffersARB(1,&vb); 
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,vb);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB,sizeof(v[0])*n,
		&v[0],GL_STATIC_DRAW_ARB);
	glVertexPointer(3,GL_FLOAT,sizeof(v[0]),
		(void *)ptrdiff(v,&v[0].loc));
	glColorPointer(3,GL_FLOAT,sizeof(v[0]),
		(void *)ptrdiff(v,&v[0].color));
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);
}

/* Draw the vertices of this mesh */
void myMesh::draw(GLenum mode)
{
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,vb);

	glDrawArrays(mode,0,len);

	glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);
}

/* Destroy the graphics card's copy of our data */
myMesh::~myMesh()
{
	glDeleteBuffersARB(1,&vb);
}

/******** Heightfield, using mesh and vertex classes *********/
/* Silly little analytic height function */
double getHeight(double x,double y) {
        return
                0.10*cos(125*(x*x+y*y))+
                0.05*cos(137*((x+0.1)*(x+0.1)+y*y));
}
vec3 makeVertex(double x,double y) {
        return vec3(x,y,getHeight(x,y));
}
vec3 makeNormal(double x,double y) {
        vec3 c=makeVertex(x,y);
        vec3 a=makeVertex(x+0.0001,y);
        vec3 b=makeVertex(x,y+0.0001);
        return (a-c).cross(b-c).dir();
}

/* Funky interference heightfield */
class myHeightfield {
public:
	std::vector<myVertex> vertices;
	myMesh *m;
	
	myHeightfield();
	void draw(void);
	~myHeightfield() {delete m;}
};


myHeightfield::myHeightfield()
{
        int n=100; /* 2*n*n triangles */
        if (0) n=3000; /* 16m triangles */
        double sz=1.0/n; /* distance between mesh vertices */
        for (int iy=0;iy<n;iy++) {
                double y=iy*sz-0.5;
                for (int ix=0;ix<n;ix++) {
                        double x=ix*sz-0.5;
                        vertices.push_back(
                                myVertex(makeVertex(x,y),makeNormal(x,y))
                        );
			/* Duplicate vertices, to split triangle strips */
			if (ix==0) vertices.push_back(vertices[vertices.size()-1]);
			
                        vertices.push_back(
                                myVertex(makeVertex(x,y+sz),makeNormal(x,y+sz))
                        );
			if (ix==n-1) vertices.push_back(vertices[vertices.size()-1]);
                }
        }
	int len=vertices.size();
	printf("Created %d vertices\n",len);
	m=new myMesh(&vertices[0],len);
}
void myHeightfield::draw(void)
{
	m->draw(GL_TRIANGLE_STRIP);
}

/* Orthonormal coordinate frame */
class ortho_frame {
public:
	/* Unit vectors pointing along axes of our frame.
		X cross Y is Z. (right-handed coordinate system)
	*/
	vec3 x,y,z;
	ortho_frame() :x(1,0,0), y(0,1,0), z(0,0,1) {}
	
	/* Reorient this coordinate frame by this far 
	   along in the X and Y axes.  "dx" is the distance
	   along the z axis to push the x axis; "dy" 
	   the distance along the z axis to push the y axis.
	*/
	void nudge(double dx,double dy) {
		x+=dx*z;
		y+=dy*z;
		orthonormalize();
	}
	
	/* Reconstruct orthonormal frame from X and Y axes.
	  Y is primary, X is secondary, Z is tertiary.
	*/
	void orthonormalize(void) {
		y=normalize(y);
		z=normalize(cross(x,y));
		x=normalize(cross(y,z));
	}
};

/* Orientation of our camera. Camera looks down -Z axis; Y is up. */
ortho_frame camera_frame;

void my_display(void) {
	glMatrixMode(GL_PROJECTION); /* set up projection matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT), /* aspect ratio: width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	vec3 camera_location=1.3*camera_frame.z; /* back away from origin */
	gluLookAt(VEC3_TO_XYZ(camera_location),
		VEC3_TO_XYZ(vec3(0)),
		VEC3_TO_XYZ(camera_frame.y));
	
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	glClearColor(0.3,0.5,0.7,0); /* set background color to blue */
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	
	glScalef(3.0,3.0,3.0);
	static myHeightfield hf;
	hf.draw();
	
	glutSwapBuffers(); /* copy back buffer to front */
	
/* Statistics */
	static int last_msec=0;
	int cur_msec=glutGet(GLUT_ELAPSED_TIME); /* milliseconds since glutInit */
	int msec_elapsed=cur_msec-last_msec;
	printf("%d msec/frame; %.2f fps\n",
		msec_elapsed,1.0e3/msec_elapsed);
	last_msec=cur_msec;
}

int mouse_x, mouse_y;
void my_mouse(int state,int buttons,int x,int y) {
	mouse_x=x; mouse_y=y;  /* just stash motion */
}
void my_motion(int x,int y) { 
	float dx=x-mouse_x, dy=y-mouse_y; /* relative motion */
	
	float speed=1.0/100; /* radians rotation per pixel */
	camera_frame.nudge(speed*dx,-speed*dy); /* OpenGL Y is negative of mouse Y! */
	
	mouse_x=x; mouse_y=y; 
	glutPostRedisplay();
}

int main(int argc,char *argv[]) {
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
        glutInitDisplayMode (mode);
	glutCreateWindow("GLSL Lighting Demo");
	glewInit();
	
	glutDisplayFunc(my_display);
	glutMotionFunc(my_motion);
	glutMouseFunc(my_mouse);
	
	glutMainLoop(); /* processes events */
	return 0;
}
